from .classification_test_cases import * 
from morebs2.modular_labeller import * 
from morebs2.onedimprt_classifier import * 
import unittest


class TestOneDimClassifierClass(unittest.TestCase):

    '''
    '''
    def test__OneDimClassifier__adjust_partition__case_1(self):
        D = OneDimClassifier_test_dataset_1() 
        L = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 
        
        # subcase 1 
        odc = OneDimClassifier(D,L,2,0)

        qx = deepcopy(odc.prt) 
        odc.adjust_partition()
        qx2 = deepcopy(odc.prt) 

        qx_ans = np.array([\
            [0.,-10.,7.55144,1.46531272],\
            [2.,-10.,7.55144,1.02932592],\
            [3.,-10.,7.55144,1.48707732],\
            [1.,-9.59191,7.55144,1.86464141],\
            [4.,-9.59191,7.55144,1.77047391]])
        qx2_ans = np.array([[2.,-10.,7.55144]])

        assert equal_iterables(qx,qx_ans) 
        assert equal_iterables(qx2,qx2_ans) 

        # subcase 2 
        odc2 = OneDimClassifier(D,L,2,1)
        qx_ = deepcopy(odc2.prt) 
        odc2.adjust_partition()
        qx2_ = deepcopy(odc2.prt) 

        qx2_ans_ = np.array([\
            [0.,-10.,1.46531272],\
            [4.,1.46531272,7.55144]])

        assert equal_iterables(qx_,qx_ans)
        assert equal_iterables(qx2_,qx2_ans_), "got {}".format(qx2_)
        return

    def test__OneDimClassifier__adjust_partition__case_2(self): 
        D = OneDimClassifier_test_dataset_1() 
        L2_ = np.array([label_vector__type_uniform_partition_sum(v,[-40,40]) for v in D])

        # subcase 1 
        odc = OneDimClassifier(D,L2_,2,0)
        qx = deepcopy(odc.prt) 
        odc.adjust_partition()
        qx2 = deepcopy(odc.prt) 

        qx_ans = np.array([\
            [0.,-10.,3.33334,-7.47449313],\
            [1.,-10.,7.55144,-5.7634775 ],\
            [2.,-10.,7.55144,0.79704803],\
            [3.,-9.58848,7.55144,5.36958762],\
            [4.,4.69136,7.55144,7.07476]])
        qx2_ans = np.array([\
            [2.,-10.,7.55144]])

        assert equal_iterables(qx,qx_ans) 
        assert equal_iterables(qx2,qx2_ans) 

        # subcase 2  
        odc = OneDimClassifier(D,L2_,2,1)
        qx = deepcopy(odc.prt) 
        odc.adjust_partition()
        qx2 = deepcopy(odc.prt) 

        qx_ans = np.array([[0.,-10.,3.33334,-7.47449313],\
            [1.,-10.,7.55144,-5.7634775],\
            [2.,-10.,7.55144,0.79704803],\
            [3.,-9.58848,7.55144,5.36958762],\
            [4.,4.69136,7.55144,7.07476]])
        qx2_ans = np.array([\
            [2.,-10.,0.79704803],\
            [3.,0.79704803,5.36958762],\
            [4.,5.36958762,7.55144]])

        assert equal_iterables(qx,qx_ans) 
        assert equal_iterables(qx2,qx2_ans) 

    def test__OneDimClassifier__adjust_partition__case_3(self): 

        D = OneDimClassifier_test_dataset_1() 
        L2__ = np.array([label_vector__type_uniform_partition_index(v,[-10,10],2,num_labels=2) for v in D])

        # subcase 1 
        odc = OneDimClassifier(D,L2__,2,0)
        
        qx = deepcopy(odc.prt) 
        odc.adjust_partition()
        qx2 = deepcopy(odc.prt) 

        qx_ans = np.array([\
            [0.,-10.,-1.11111,-6.19388667],\
            [1.,1.11111,7.55144,4.87319363]])
        qx2_ans = np.array([\
            [0.,-10.,-1.11111],\
            [1.,-1.11111,7.55144]])

        assert equal_iterables(qx,qx_ans) 
        assert equal_iterables(qx2,qx2_ans), "got \n{} \nexpected \n{}".format(qx2,qx2_ans)

        # subcase 2 
        odc = OneDimClassifier(D,L2__,2,1)

        qx = deepcopy(odc.prt) 
        odc.adjust_partition()
        qx2 = deepcopy(odc.prt) 

        qx_ans = np.array([\
            [0.,-10.,-1.11111,-6.19388667],\
            [1.,1.11111,7.55144,4.87319363]])

        qx2_ans = np.array([\
            [0.,-10.,-6.19388667],\
            [1.,-6.19388667,7.55144]])

        assert equal_iterables(qx,qx_ans) 
        assert equal_iterables(qx2,qx2_ans) 

    def test__OneDimClassifier__adjust_partition__case_4(self): 


        D2 = np.zeros((1000,4)) 
        D2[:,0] += 10 
        D2[:,1] -= 5
        D2[:,2] += 10 
        D2[:,3] -= 5 

        D2[:500,0] /= 2 
        D2[500:,1] *= 2 
    
        L2 = [0] * 500 + [1] * 500 
        L2 = np.array(L2)

        odc = OneDimClassifier(D2,L2,0,0) 
        odc.make() 
        c3 = 0 
        for d,l in zip(D2,L2): 
            l_ = odc.classify(d)
            c3 += bool(l_ == l)
        assert c3 == 1000, "got {}".format(c3) 

class TestRecursiveOneDimClassifierClass(unittest.TestCase): 

    def test__RecursiveOneDimClassifier__fit_case_1(self): 

        D = OneDimClassifier_test_dataset_1() 
        L2 = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 

        # subcase 1
        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        #prg = prg__LCG(0,2,1,21) 
        rodc = RecursiveOneDimClassifier(D,L2,prg,0)
        rodc.fit() 
        c = rodc.score_accuracy(D,L2)

        # subcase 2
        rodc2 = RecursiveOneDimClassifier(D,L2,prg,1)
        rodc2.fit() 
        c2 = rodc2.score_accuracy(D,L2)

        # subcase 3 
        rodc3 = RecursiveOneDimClassifier(D,L2,prg,prg)
        rodc3.fit() 
        c3 = rodc3.score_accuracy(D,L2)

        assert c == 238
        assert c2 == 283
        assert c3 == 210, "got {}".format(c3)

    def test__RecursiveOneDimClassifier__fit_case_2(self): 
        D = OneDimClassifier_test_dataset_1() 
        L2_ = np.array([label_vector__type_uniform_partition_sum(v,[-40,40]) for v in D])

        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 

        # subcase 1
        rodc = RecursiveOneDimClassifier(D,L2_,prg,0)
        rodc.fit() 
        c = rodc.score_accuracy(D,L2_)

        # subcase 2
        rodc2 = RecursiveOneDimClassifier(D,L2_,prg,1)
        rodc2.fit() 
        c2 = rodc2.score_accuracy(D,L2_)

        # subcase 3
        rodc3 = RecursiveOneDimClassifier(D,L2_,prg,prg)
        rodc3.fit() 
        c3 = rodc3.score_accuracy(D,L2_)

        assert c == 563, "got {}".format(c)
        assert c2 == 760, "got {}".format(c2)
        assert c3 == 459, "got {}".format(c3)

    def test__RecursiveOneDimClassifier__fit_case_3(self): 
        D = OneDimClassifier_test_dataset_1() 
        L2__ = np.array([label_vector__type_uniform_partition_index(v,[-10,10],2,num_labels=2) for v in D])

        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 

        # subcase 1
        rodc = RecursiveOneDimClassifier(D,L2__,prg,0)
        rodc.fit() 
        c = rodc.score_accuracy(D,L2__)

        # subcase 2
        rodc2 = RecursiveOneDimClassifier(D,L2__,prg,1)
        rodc2.fit() 
        c2 = rodc2.score_accuracy(D,L2__)

        # subcase 3
        rodc3 = RecursiveOneDimClassifier(D,L2__,prg,prg)
        rodc3.fit() 
        c3 = rodc3.score_accuracy(D,L2__)

        assert c == 1000
        assert c2 == 862
        assert c3 == 719, "got {}".format(c3)

    def test__RecursiveOneDimClassifier__fit_case_4(self): 
        D = np.zeros((1000,4)) 
        D[:,0] += 10 
        D[:,1] -= 5
        D[:,2] += 10 
        D[:,3] -= 5 

        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        f = lambda x,x2: x - x2
        L = np.array([label_vector__type_prgANDindex_output(i,prg,f,5) for i in range(len(D))]) 

        rodc = RecursiveOneDimClassifier(D,L,prg,0)
        rodc.fit() 
        c = rodc.score_accuracy(D,L)
        assert c == 199 

    def test__RecursiveOneDimClassifier__fit_case_5(self): 
        # subcase 1 
        D2 = np.zeros((1000,4)) 
        D2[:,0] += 10 
        D2[:,1] -= 5
        D2[:,2] += 10 
        D2[:,3] -= 5 

        D2[:500,0] /= 2 
        D2[500:,1] *= 2 

        L2 = [0] * 500 + [1] * 500 
        L2 = np.array(L2)

        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        
        rodc = RecursiveOneDimClassifier(D2,L2,prg,0)
        rodc.fit() 
        c = rodc.score_accuracy(D2,L2)
        assert c == 1000, "got {}".format(c) 

        # subcase 2 
        D3 = deepcopy(D2) 
        D3[250:500,2] *= 2 
        D3[250:500,3] /= 2 
        L3 = [0] * 250 + [1] * 250 + [2] * 250 + [3] * 250
        L3 = np.array(L3) 

        rodc2 = RecursiveOneDimClassifier(D3,L3,prg,0)
        rodc2.fit() 
        c2 = rodc2.score_accuracy(D3,L3)
        assert c2 == 750, "got {}".format(c2)  

    

if __name__ == "__main__":
    unittest.main()



from .onedimprt_classifier import * 
from .seq_repr import contiguous_repr__sequence

"""
First node (root) for classifier<AdditiveAdjustedRecursiveODC>. 
Adds a vector v_l to input x, according to current index i that maps 
to vector v_l corresponding to a label l. 
"""
class ODCAdditiveAdjustmentNode: 

    def __init__(self,contiguous_sequence,additive_map):  
        self.contiguous_sequence = contiguous_sequence
        self.additive_map = additive_map 

        # counter variable
        self.c = 0
        self.index_counter = 0 
        self.part_index = 0 

    def map(self,v): 
        self.c += 1 
        self.index_counter += 1 

        q = self.contiguous_sequence[self.part_index] 
        # case: move to next 
        if self.index_counter > q[1]: 
           #print("NEXT PART FROM ",q) 
            self.index_counter = 1 
            self.part_index += 1 
            if self.part_index >= len(self.contiguous_sequence): 
                self.part_index = 0

        q = self.contiguous_sequence[self.part_index] 
        l = q[0] 
        v2 = self.additive_map[l] 
        return v + v2

    def reset(self): 
        self.c = 0 
        self.index_counter = 0 
        self.part_index = 0 

"""
A memory-intensive extension to classifier<RecursiveOneDimClassifier>. 
Practically memorizes all (index,label) pairs for label vector L. 
With this information, classifier uses its root node to add the appropriate 
vector v_l to every input v, before classifier<RecursiveOneDimClassifier> 
proceeds with classifying (v_l + v). 
"""
class AdditiveAdjustedRecursiveODC(RecursiveOneDimClassifier): 

    def __init__(self,D,L,prg=None,pscheme=0,verbose:bool=False): 
        D = self.preproc(D,L)
        super().__init__(D,L,prg,pscheme,verbose) 

    #--------------------------------------------------------------------

    def preproc(self,D,L): 
        self.make_additive_map(D,L) 
        self.make_adjustment(L) 
        self.root_ = ODCAdditiveAdjustmentNode(self.crepr,self.add_map) 

        D2 = np.array([self.root_.map(d) for d in D]) 
        self.root_.reset() 
        return D2 

    def make_adjustment(self,L): 
        self.crepr = contiguous_repr__sequence(L)
        
    def make_additive_map(self,D,L):
        labels = sorted(set(L)) 
        num_cols = D.shape[1]
        max_diff = max([max(D[:,i]) - min(D[:,i]) for i in range(num_cols)]) 
        increment = max_diff * 2  

        if increment == 0: 
            increment = 10 

        v = np.zeros((num_cols,)) 
        add_map = dict() 

        for (i,l) in enumerate(labels): 
            v_ = np.copy(v) 
            v_ +=  increment + increment * i * -1 ** i 
            add_map[l] = v_ 
        self.add_map = add_map 

    #----------------------------------------------------------------------

    """
    main method #2 
    """
    def classify(self,v): 
        v2 = self.root_.map(v) 
        return super().classify(v2) 

    """
    main method #3 
    """
    def score_accuracy(self,D,L): 
        assert len(D) == len(L) 

        c = 0 
        for (x,l) in zip(D,L): 
            l_ = self.classify(x) 
            if l_ == l: 
                c += 1 
        return c 

#------------------------------------------------------------------------------------------------



from .classification_test_cases import * 
from morebs2.modular_labeller import * 
from morebs2.odc_extended import * 
import unittest

'''
py -m morebs2.tests.odc_extended_test  
'''
class AdditiveAdjustedRecursiveODCClass(unittest.TestCase):

    def test__AdditiveAdjustedRecursiveODC__score_accuracy_case1(self):

        D = OneDimClassifier_test_dataset_1() 
        L = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 

        # subcase 1
        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        rodc = AdditiveAdjustedRecursiveODC(D,L,prg,0)
        rodc.fit() 

        c = rodc.score_accuracy(D,L)
        c2 = rodc.score_accuracy(D,L)
        assert c == c2 == 1000, "got {}".format(c)

        # subcase 2 
        rodc = AdditiveAdjustedRecursiveODC(D,L,prg,1)
        rodc.fit() 

        c = rodc.score_accuracy(D,L)
        c2 = rodc.score_accuracy(D,L)
        assert c == c2 == 904, "got {}".format(c)


        return

if __name__ == '__main__':
    unittest.main()


from morebs2.deline_mc import * 
#from collections import Counter 
#import numpy as np
import unittest
import time 

def DLineateMCApproximators_sample_dataset(): 

    prg = prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545) 
    def prg0(): 
        return modulo_in_range(prg(),[-10.,10]) 

    def prg1(): 
        return modulo_in_range(prg(),[5,15]) 

    def prg2(): 
        return modulo_in_range(prg(),[12,32])  

    pv0 = prg__single_to_nvec(prg0,2) 
    pv1 = prg__single_to_nvec(prg1,2) 
    pv2 = prg__single_to_nvec(prg2,2) 

    num_samples = 2000 
    xy0 = [pv0() for _ in range(num_samples)] 
    xy1 = [pv1() for _ in range(num_samples)] 
    xy2 = [pv2() for _ in range(num_samples)] 
    xy = np.array(xy0 + xy1 + xy2)
    L = [0 for _ in range(num_samples)] 
    L1 = [1 for _ in range(num_samples)] 
    L2 = [2 for _ in range(num_samples)] 
    L = np.array(L + L1 + L2).reshape((num_samples *3,1))
    xyl = np.hstack((xy,L))
    return xyl 

'''
py -m morebs2.tests.deline_mc_test  
'''
class DLineateMCApproxClass(unittest.TestCase):

    def test__DLineateMCApprox__one_classification__case_1(self): 
        xyl = DLineateMCApproximators_sample_dataset() 
        d1 = DLineateMCApprox(xyl,dmethod="nodup",prg=prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545))
        d1.full_run() 

        c = 0 
        for (i,x) in enumerate(xyl): 
            x2 = x[:2]
            rx = d1.one_classification(x2)
            #print("i: ",i, " label: ",x[2], " got: ",rx)
            c += int(x[2] == rx) 
        assert c == 1396

class DLineateMCApproximatorsClass(unittest.TestCase):

    """
    total runtime on personal device: ~ 205 seconds. 
    runtime to find classifier: ~ 40 seconds.
    runtime to classify: ~ (205 - 40) 
    number of samples: 6K 
    """
    def test__DLineateMCApproximators__one_classification__case_1(self):
        xyl = DLineateMCApproximators_sample_dataset() 
        d1 = DLineateMCApproximators(xyl,20,dmethod="nodup",prg=prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545))

        t = time.time() 
        d1.full_run() 
        t2  = time.time() 
        print("search runtime: ",t2 - t)
        c = 0 
        for (i,x) in enumerate(xyl): 
            x2 = x[:2]
            rx = d1.one_classification(x2)
            #print("i: ",i, " label: ",x[2], " got: ",rx)
            c += int(x[2] == rx) 

        print("classification runtime: ",time.time() - t2)


if __name__ == "__main__":
    unittest.main()



from morebs2.ball_comp import *
from morebs2.message_streamer import *
from morebs2.ball_comp_test_cases import *
import unittest    

'''
py -m morebs2.tests.ball_comp_test  
'''
class BallCompClass(unittest.TestCase):

    def test__BallComp__conduct_decision__sample_data_1__case1(self):
        maxBalls = 20
        maxRadius = 5.0
        td = ballcomp_sample_data_1()
        bc = BallComp(maxBalls,maxRadius,None,True)
        
        S = [] 
        for t in td:
            s = bc.conduct_decision(t)
            S.append(s) 

        assert S == [2, 2, 2, 2, 2, 2, 2, 2]
        assert len(bc.balls) == 8 
        return

    def test__BallComp__conduct_decision__sample_data_1__case2(self):

        maxBalls = 4 
        maxRadius = 8.0

        td = ballcomp_sample_data_1()
        bc = BallComp(maxBalls,maxRadius,None,False)

        S = [] 
        L = [] 
        for (i,t) in enumerate(td):
            s = bc.conduct_decision(t)
            S.append(s) 

            l = len(bc.balls) 
            L.append(l)

            if i >= 5: 
                assert bc.terminateDelta
            
        assert S == [2, 2, 2, 1, 2, -1, -1, -1],"got {}".format(S) 
        assert L == [1, 2, 3, 3, 4, 4, 4, 4] 
        
        ball_info = {0:(1,5),\
                    1:(1,5),\
                    3:(2,5),\
                    4:(1,5)} 

        Q = bc.summarize_ball_info()
        for q in Q: 
            ans = ball_info[q[0]] 
            assert ans == q[2] 

    def test__BallComp__conduct_decision__sample_data_2(self): 

        maxBalls = 4
        maxRadius = 0.2
        td = ballcomp_sample_data_2()
        vh = ViolationHandler1(maxBalls,maxRadius)
        bc = BallComp(maxBalls,maxRadius,vh,0)#2)

        c = None 
        for i,t in enumerate(td):
            s = bc.conduct_decision(t)
            print("point size: ",bc.point_size(), s,bc.terminateDelta)
            if type(c) == type(None) and s == -1: 
                c = i 

        print("********************")
        print("BALLS ", len(bc.balls))
        for k,v in bc.balls.items():
            print("k ",k)
            print(v)
            print()
        assert c == 5 

    def test__BallComp__conduct_decision__sample_data_3(self): 

        maxBalls = 5
        maxRadius = .5
        td = ballcomp_sample_data_3()
        vh = ViolationHandler1(maxBalls,maxRadius)

        # TODO: delete k
        bc = BallComp(maxBalls,maxRadius,vh,True)
        print("TD ",td.shape)

        c = None 
        L = []  
        for (i,t) in enumerate(td):
            s = bc.conduct_decision(t)
            if type(c) == type(None) and s == -1:
                c = i 
            L.append(s) 

        assert L == [2, 2, 2, 2, 1, 2, -1, -1] 
        assert c == 6 

        print("********************")
        print("BALLS ", len(bc.balls))
        for k,v in bc.balls.items():
            print("k ",k)
            print(v)
            print()

    def test__BallComp__conduct_decision__sample_data_4(self): 

        maxBalls = 5
        maxRadius = 20.0

        vh = ViolationHandler1(15,80.0)

        filePath = "morebs2/indep/ballcomp_sample_data_4.txt"
        ms = MessageStreamer(filePath,readMode = 'r')
        bc = BallComp(maxBalls,maxRadius,vh,0)

        q = 20
        s = 0
        S = [] 
        while ms.stream() and q > 0:
            for t in ms.blockData:
                d = bc.conduct_decision(t)
                if d != -1:
                    s += 1
                S.append(d) 
            q -= 1

        assert S.count(1) == 1254 and S.count(2) == 6 
        assert bc.point_size() == 2493 and s == 1260
        ms.end_stream() 

    def test__BallComp__conduct_decision__sample_data_5(self): 
        D = ballcomp_sample_data_LCG_1() 

        num_balls = 50 
        max_radius = 2000  
        vh = ViolationHandler1(num_balls,max_radius)
        bc = BallComp(num_balls,max_radius,vh,verbose=0)

        S = [] 
        L0 = [] 
        for v in D: 
            s = bc.conduct_decision(v) 
            S.append(s)
            L0.append(bc.ball_label_for_point(v)) 
        S = np.array(S)

        L1 = [] 
        L1_ = [] 
        for v in D: 
            L1_.append(bc.ball_label_for_point__qualify_radius(v))  
            L1.append(bc.ball_label_for_point(v)) 
        L1 = np.array(L1)

        freq_1 = list(S).count(1)
        freq_2 = list(S).count(2)
        assert freq_1 == 1652 and freq_2 == 348 

        print("check cumulative difference between real-time and post-solution labeling") 
        assert np.sum(np.abs(L1 - L0)) == 30137 


if __name__ == "__main__":
    unittest.main()